function [strav, bopt, strav_t, theta, pphi, y_aligned] = sim_estimate(t, x, do_polyfit, polyord, niter, effective_time_method, do_norm)

% [strav, bopt, strav_t, theta, pphi, y_aligned] = sim_estimate(t, x, do_polyfit, polyord, niter, effective_time_method, do_norm)
%
% Estimate the SIM model (Shape Invariant Model) for some curves using the
% iterative algorithm in the paper:
%
% A.Kneip and J.Engel. "Model Estimation in Nonlinear Regression Under
% Shape Invariance". The Annals of Statistics, vol.23, no.2 (April 1995),
% pp.551-570.
%
%
% The SIM model represents each curve as:
%
% f_i(t) = a_i * g( (t-c_i)/b_i ) + d_i
%
% where g(t) is called "the structural average curve".
%
%---------
% Input:
%
% t:            time vector of each curve, n*1 or 1*n
%
% x:            curve data, m*n matrix (m: # curves)
%
% do_polyfit:   whether do polyfit for extrapolation, [optional, default 1]
%
% polyord:      if do polyfit, order of the polynomial, [optional, default 6]
%
% niter:        iteration number [optional, default 5]
%
% effective_time_method: 
%               method to determine the time range for the curve g(t), strav_t
%               0 - defalut, the extrapolated range (>=[0,1])
%               1 - the overlapped range (<=[0,1]) This is the method in the
%                   original paper above.
%               2 - the same as the input time vector (=[0,1]).
%
% do_norm:      whether do normalization, [optional, default 1]
%
%---------
% Output:
%
% strav:        structural average curve, n_strav*1 vector
%               and also, mean(y_aligned) == strav. (n_strav=100)
%
% bopt:         the optimal bandwidth for strav
%               Note that using (strav, bopt, t), we can get the
%               kernel-smoothed function of the structural average curve. For,
%               example, if we need to get the function value in time vector
%               tt, we can call:
%
%                   yy = smodtm(t, strav', tt, 0, bopt);
%
% strav_t:      the effective time of the structural average curve. It should
%               be greater or equal to the range of the original time vector,
%               t. Its formula [t1,t2] depends on effective_time_method above.
%           
%               effective_time_method = 0 (default):
%                   t1 = min((t(1)-theta(:,3))./theta(:,2))
%                   t2 = max((t(n)-theta(:,3))./theta(:,2))
%
%               effective_time_method = 1 (the paper proposed one):
%                   t1 = max((t(1)-theta(:,3))./theta(:,2))
%                   t2 = min((t(n)-theta(:,3))./theta(:,2))
%
%               effective_time_method = 2:
%                   t1 = t(1)
%                   t2 = t(n)
%
% theta:        the parameters of SIM model for each curve
%               m*4 matrix. Each rows is (a_i, b_i, c_i, d_i)
%
% pphi:         the polyfit result of strav 
%
% y_aligned:   
%               the aligned curve dat. m*n matrix. Each row is given by:
%
%                   a_i * g((t-c_i)/d_i) + d_i
%
%               It can be used to compute the reconstruction error with the
%               original data f_i(t).            
%
%
%
% See also Kd_sim_estimate.m, smodtm.c, glkern.c
%
%
%---------
% Comments:
%
% Experimental results show that, for the best data fitting, there are two
% configurations:
%
% 1. use polyfit after each iteration to ensure the extra-interpolation of the strav.
%    For the final output, however, still use kernel-based interpolation. 
%    This is the default setting and works well for various curves.
% 
% 2. The original method proposed in the above paper is not using polyfit. 
%    By experiments, we found this method can also fit 
%    the data well. But the strav has limited range. That is, the extra-interpolation ability 
%    is weak. For example, it can not fit well for the saturated part of the S-shape curves. 
%
%
% Jinwei Gu. 2005/12/06
% jwgu@cs.columbia.edu

if nargin < 3
    do_polyfit = 1;
end

if nargin < 4
    polyord = 6;
end

if nargin < 5
    niter = 5;
end

if nargin < 6
    effective_time_method = 0;
end

if nargin < 7
    do_norm = 1;
end

[m, n] = size(x);
if length(t)~=n
    error('t and x must have the same length.');
end
theta = zeros(m, 4);

%-------------------------------------------------------
%%% Step 1:
% initialization
% the simplest initialization

theta(:,1) = 1;     
theta(:,2) = 1;
theta(:,3) = 0;
theta(:,4) = 0;

% initialization step in the paper
% TODO: implement the more robust and accurate 
%       initialization method in the paper.
% ....

%-------------------------------------------------------
%%% Step 2:
% initialization of \phi(t) strav, effective time strav_t,
% and the optimal band width b.

[y, b]  = smodtm(t, x, t);
b       = mean(b);
n_strav = 100; % strav curve is represented as 100 points

a0 = (t(1)-theta(:,3))./theta(:,2);
a1 = (t(n)-theta(:,3))./theta(:,2);    
if effective_time_method==0
    strav_t = linspace(min(a0), max(a1), n_strav);
elseif effective_time_method==1
    strav_t = linspace(max(a0), min(a1), n_strav);
else
    strav_t = linspace(t(1), t(n), n_strav);
end

y = zeros(m, n_strav);
for i=1:m
    y(i,:) = smodtm(t, x(i,:), theta(i,2)*strav_t+theta(i,3), 0, b);
end    
	
for i = 1:n_strav
    idx    = find(and(a0<=strav_t(i), strav_t(i)<=a1));                        
    phi(i) = (sum(y(idx,i))-sum(theta(idx,4)))/sum(theta(idx,1));
end

pphi = 0;
if do_polyfit == 1
    pphi = polyfit(strav_t, phi, polyord);
end

%-------------------------------------------------------
%%% Step 3:
% Two-step optimization algorithm to estimate
% strav, strav_t, and theta

options = optimset('Display', 'Off');
for iter = 1:niter
    
    % estimate the optimal theta for each curve
    errs = 0; nconverge = 0;
    for i = 1:m       
        
        if do_polyfit == 0
            [p, err, res, flag] = lsqnonlin(@myfun, theta(i,:), [0,0,-inf,-inf], [inf,inf,inf,inf], ...
                                            options, t, x(i,:), b, phi, strav_t);
        else
            [p, err, res, flag] = lsqnonlin(@myfun_polyfit, theta(i,:), [0,0,-inf,-inf], [inf,inf,inf,inf], ...
                                            options, t, x(i,:), b, pphi);
        end

        theta(i,:) = p;
        errs = errs + err;
        
        if flag > 0
            nconverge = nconverge + 1;
        end
    end
    fprintf('iter %d\terr %f\tconverge: %f\n', iter, err, nconverge/m);
        
    if do_norm == 1 % normalize theta        
        theta(:,1) = m*theta(:,1)/sum(theta(:,1));
        theta(:,2) = m*theta(:,2)/sum(theta(:,2));
        theta(:,3) = theta(:,3) - sum(theta(:,3))*theta(:,2)/m;
        theta(:,4) = theta(:,4) - sum(theta(:,4))*theta(:,1)/m;
    end
    
    % compute the \phi(t) strav, and strav_t from the estimate theta
    a0 = (t(1)-theta(:,3))./theta(:,2);
    a1 = (t(n)-theta(:,3))./theta(:,2); 
    if effective_time_method==0
        strav_t = linspace(min(a0), max(a1), n_strav);
	elseif effective_time_method==1
        strav_t = linspace(max(a0), min(a1), n_strav);
	else
        strav_t = linspace(t(1), t(n), n_strav);
    end
	
    % for some samples, more iterations after converged seems harmful for
    % fitting error. Stop iteration when almost all converged.
    if nconverge/m >= 0.99
        break;
    end
    
    for i=1:m
        y(i,:) = smodtm(t, x(i,:), theta(i,2)*strav_t+theta(i,3), 0, b);
    end    
		
    for i = 1:n_strav
        idx    = find(and(a0<=strav_t(i), strav_t(i)<=a1));                        
        phi(i) = (sum(y(idx,i))-sum(theta(idx,4)))/sum(theta(idx,1));
    end
            
    if do_polyfit == 1
        pphi = polyfit(strav_t, phi, polyord);
    end
end

kord = 2; % order of the kernel function 
bopt = m^(-1/(2*kord+1)) * b; 

for i=1:m
    y(i,:) = smodtm(t, x(i,:), theta(i,2)*strav_t+theta(i,3), 0, bopt);
end

for i = 1:n_strav
    idx      = find(and(a0<=strav_t(i), strav_t(i)<=a1));                        
    strav(i) = (sum(y(idx,i))-sum(theta(idx,4)))/sum(theta(idx,1));
end
    
if do_polyfit == 1
    pphi = polyfit(strav_t, strav, polyord);
end

if nargout==6   
     % Experimental results show that even using polyfit for
     % extrainterpolation, for the final output, the kernel-smoothed method
     % is still the best fit for the original data
     %
     % Jinwei Gu. 2005/12/05    
    for i = 1:m
        z = smodtm(strav_t, strav, (t-theta(i,3))/theta(i,2), 0, bopt);
        y_aligned(i,:) = theta(i,1)*z + theta(i,4);
    end
end

%---------------------------------------------
function f = myfun(p, t, x, b, phi, strav_t)

z = smodtm(strav_t, phi, (t-p(3))/p(2), 0, b);
f = x - p(1)*z - p(4);
f = f';

%---------------------------------------------
function f = myfun_polyfit(p, t, x, b, pphi)

z = polyval(pphi, (t-p(3))/p(2));
f = x - p(1)*z - p(4);
f = f';
